# To compile: gcc -O3 -shared -o hensel.so -fPIC hensel.c polyutil9.c

read "CMSHL_NMContLI3.mpl";
read "MakeBlackBoxes.mpl";

CCodeDir := "./";

MUL64S := define_external('mul64s',
  AA::integer[8], BB::integer[8], pp::integer[8], LIB=cat(CCodeDir,"detinterp4.so"),
  RETURN::integer[8]):
  MUL64S := subsop(1=(AA,BB,pp),eval(MUL64S)):

  INV64S := define_external('inv64s',
  AA::integer[8], pp::integer[8], LIB=cat(CCodeDir,"detinterp4.so"),
  RETURN::integer[8]):
  INV64S := subsop(1=(AA,pp),eval(INV64S)):

MonCoffs := proc(f, X::list) local m;
    return [coeffs(f, X, 'm')], [m];
end:

Bivarpoly2matrix := proc( a, X::list, M::Array )
    # Input: a in Zp[x,y], X=[x,y], d = [deg(a,x),deg(a,y)]
    # Output: M with A[i,j] = coefficient of x^j*y^i in a, i=0..dy, j=0..dx
    local cof,mon,dm1,dm2,k,dx,dy,x:=X[1],y:=X[2];
    dx := degree(a,x); dy := degree(a,y); #print(a); print(X);
    cof,mon := MonCoffs(a,X);
    dm1 := map(degree,mon,x); dm2 := map(degree,mon,y);
    for k to nops(mon) do M[dm2[k],dm1[k]] := cof[k]; od;
end:


initialf2matrix := proc( f0::list, x, F0::Array )
    # Input: f0 is a list of factors in Zp[x] s.t. a(y=alpha) = mul(f0[i],i=1..r) mod p, X=[x,y]
    # Ouput: Matrix F0 s.t. row rr is the array of coefficients of f0[rr] for rr=1..r
    local r:=nops(f0), rr, i, df0;
    df0 := map(degree,f0,x);
    for rr to r do
        for i from 0 to df0[rr] do F0[rr,i] := coeff(f0[rr],x,i); od;
    od;
    return;
end:

TdegEstimate := proc(BB,X,p)

   local i,j,k,xx,yy,ra,n,beta,fest1,fest2,count,alpha,gamma;

   n := nops(X);

   ra := rand(1..p-1);
   beta := [seq(ra(),i=1..n)];
   fest1 := 0; fest2 := 1; count := 0;

   xx := []; yy := [];

   while fest1 <> fest2 or count < 3 do
      alpha := ra();
     while evalb(alpha in xx) do alpha := ra(); od;
      xx := [op(xx),alpha];
      yy := [op(yy),BB(alpha*beta,p)];
      fest1 := Interp(xx,yy,x) mod p;
      if fest1 = fest2 then count := count+1;
      else  count := 0; fi;
      fest2 := fest1;
   od;

   return degree(fest1,x);

end proc:

ProbeInterpUni := proc(A,X,x,j,p)

   local i, m, ra, Y, g,g2,f,xPoints,yPoints, check;

   ra := rand(1..p-1);
   m := nops(X);

   check := 0;
   g := -1;
   Y := X;

   Y[j] := ra();
   xPoints := [Y[j]]; yPoints := [A(Y,p)];
   g2 := Interp(xPoints,yPoints,x) mod p;

   while g <> g2 and check < 4 do
      g := g2; #print(g2);
      while evalb(Y[j] in xPoints) do
         Y[j] := ra();
      od;
      xPoints := [op(xPoints),Y[j]]; yPoints := [op(yPoints),A(Y,p)];
      g2 := Interp(xPoints,yPoints,x) mod p;
      if g = g2 then check = check + 1;
      else check := 0; fi;
   od;

   return g;

end proc:

ProbeInterpUni2 := proc(BA,X,alpha,x,j,p)

   local i, n, ra, b,xPoints,yPoints,beta,k,check,alphak;

   beta := alpha;
   ra := rand(1..p-1);
   n := nops(X);

   b[-1] := -1; b[0] := 0;
   xPoints := []; yPoints := [];
   k := -1; check := 0;

   while b[k-1] <> b[k] and check < 3 do
      k := k+1;
      alphak := ra();

      while evalb(alpha[j]*alphak in xPoints) do
         alphak := ra();
      od;

      beta[j] := alphak*alpha[j] mod p;
       xPoints := [op(xPoints),beta[j]];
      yPoints := [op(yPoints),BA(beta,p)];

      b[k] := Interp(xPoints,yPoints,x) mod p; #print(b[k]);

      if b[k-1] = b[k] then
         check := check + 1;
      else
         check := 0;
      fi;
   od;

   return b[k];

end proc:

UnivInterpNew := proc(B,X,alpha2,j,p)

   local n,ra,beta,i,M,b,alpha,k,alphak,yk,ba,Ma,vk;

   n := nops(X);

   if j < 1 or j > n then
      return "FAIL";
   fi;

   ra := rand(1..p-1);
   beta := alpha2;
   alpha := Array(1..10);


   M := 1; b[-1] := 0; k := -1;
   do
     k := k+1;
     alphak := ra();
     if evalb(alphak in alpha) then
        alphak := ra();
     fi;
     if k+1 > numelems(alpha) then
        alpha := ArrayTools[Extend](alpha,[seq(0,i=1..numelems(alpha))]);
     fi;
     #print(alpha);
     alpha[k+1] := alphak;
     beta[j] := alphak;
     yk := B(beta,p); #print(yk);
     ba := Eval(b[k-1],x=alphak) mod p;
     Ma := Eval(M,x=alphak) mod p;
     vk := (yk - ba)/Ma mod p;
     b[k] := b[k-1] + vk*M mod p; #print(b[k]);
     M := M*(x-alphak); #print(degree(M));
    # print(b[k]);
     #print(b[k]);

   until vk = 0;


   return b[k];

end proc:
TotalDegNew := proc(BA,X,p)

   local n,ra,beta,i,j,M,b,alpha,k,alphak,yk,ba,Ma,vk,beta2;

   n := nops(X);

   ra := rand(1..p-1);
   beta := [seq(ra(),i=1..n)];
   alpha := Array(1..10);


   M := 1; b[-1] := 0; k := -1;
   do
     k := k+1;
     alphak := ra();
     if evalb(alphak in alpha) then
        alphak := ra();
     fi;
     if k+1 > numelems(alpha) then
        alpha := ArrayTools[Extend](alpha,[seq(0,i=1..numelems(alpha))]);
     fi;
     #print(alpha);
     alpha[k+1] := alphak;
     beta2 := [seq(beta[j]*alphak mod p,j=1..n)];
     yk := BA(beta2,p);
     ba := Eval(b[k-1],x=alphak) mod p;
     Ma := Eval(M,x=alphak) mod p;
     vk := (yk - ba)/Ma mod p;
     b[k] := b[k-1] + vk*M mod p;
     M := M*(x-alphak);
     #print(b[k]);

   until vk = 0;


   return degree(b[k]);

end proc:


ProbeInterpDegUni := proc(BB,X,x,j,deg,alpha,p)

   local i,xpoints,ypoints,t,ra,f,beta;

   beta := alpha;

   if p-1 < deg+1 then
      printf("p is too small to perform this interpolation");
      return FAIL;
   fi;


   ra := rand(1..p-1);

   xpoints := [seq(0,i=1..deg+1)]; ypoints := [seq(0,i=1..deg+1)];
   #print(xpoints); print(ypoints);

   #fill xpoints with unique elements of Zp
   i := 1;
   while i <= deg+1 do
      t := ra();
      if not evalb(t in xpoints) then
         xpoints[i] := t;
         i := i+1;
      fi;
   od;

#print(xpoints);

   for i from 1 to deg+1 do
      beta[j] := xpoints[i];
      ypoints[i] := BB(beta,p);
   od;

   f := Interp(xpoints,ypoints,x) mod p;

   return f;

end proc:
ProbeInterpDegBi := proc(BB,X,x,j,y,k,degx,degy,alpha,p)

   local i,m,xpointsuni,ypointsuni,xpointsbi,ypointsbi,t,ra,f,beta;

   beta := alpha;

   if p-1 < degx+1 or p-1 < degy+1 then
      printf("p is too small to perform this interpolation");
      return FAIL;
   fi;

   ra := rand(1..p-1);

   xpointsuni := [seq(0,i=1..degx+1)]; ypointsuni := [seq(0,i=1..degx+1)];
   xpointsbi := [seq(0,i=1..degy+1)]; ypointsbi := [seq(0,i=1..degy+1)];

   #fill xpoints with unique elements of Zp
   i := 1;
   while i <= degx+1 do
      t := ra();
      if not evalb(t in xpointsuni) then
         xpointsuni[i] := t;
         i := i+1;
      fi;
   od;
   i := 1;
   while i <= degy+1 do
      t := ra();
      if not evalb(t in xpointsbi) then
         xpointsbi[i] := t;
         i := i+1;
      fi;
   od;

   for i from 1 to degy+1 do
      beta[k] := xpointsbi[i];
      for m from 1 to degx+1 do
         beta[j] := xpointsuni[m];
         ypointsuni[m] := BB(beta,p);
      od;
      ypointsbi[i] := Interp(xpointsuni,ypointsuni,x) mod p;
   od;

   f := Interp(xpointsbi,ypointsbi,y) mod p;

   return f;

end proc:


with(LinearAlgebra):

#check viable alpha for use in Gcd() mod p
checkAlpha := proc(A,B,X,Alpha,p)

   local Amod, Bmod, viableAB, badAlphaCounter, ra, a, b, g, alpha, LCA, LCB,x,y,abar,bbar;

   x := X[1]; y := X[2];

   Amod := A; Bmod := B; alpha := Alpha[2];
   viableAB := 0;

   #print("check alpha");
   #print(A); #print(B);

   #print(lcoeff(A,x)); #print(lcoeff(B,x));

   ra := rand(0..p-1);

   #make sure we have a viable alpha
   LCA := lcoeff(A,x); LCB := lcoeff(B,x);
   while Eval(LCA,y=alpha) mod p = 0 or Eval(LCB,y=alpha) mod p = 0 do
      alpha := ra();
   od;

   #print(alpha);

   #find g such that gcd(g,abar)=1 and gcd(g,bbar)=1
    while viableAB < 2 do

	    ##print(alpha);

        viableAB := 0;

        a := Eval(Amod,y=alpha) mod p;
        b := Eval(Bmod,y=alpha) mod p;

        g := Gcd(a,b,'abar','bbar') mod p;


        if Gcd(g,abar) mod p = 1 then
           viableAB := viableAB + 1;
        else
           Amod := Expand(A + ra()*B) mod p;
        fi;

        if Gcd(g,bbar) mod p = 1 then
           viableAB := viableAB + 1;
        else
           Bmod := Expand(A + ra()*B) mod p;
        fi;

    od;

    return Amod, Bmod, g, abar, bbar, alpha;

end proc:


UNPACK := proc(f)

   local co,terms,factors,multip,i,r;

   co := f[1]; terms := f[2];
   r := nops(terms);
   factors := [seq(0,i=1..r)];
   multip := [seq(0,i=1..r)];
   for i from 1 to r do
      factors[i] := terms[i][1];
      multip[i] := terms[i][2];
   od;

   return r,factors,multip;
end proc:


local uniDegTime := time() - time();
local totDegTime := time() - time();
local uniGcdTime := time() - time();
local CMBBSHLTime := time() - time();
local buildTime := time() - time();
local checkTime := time() - time();
local gcdPTime := time() - time();
local sqfrTime := time() - time();
local chremTime := time() - time();
local expandTime := time() - time();

resetTimes := proc()

uniDegTime := time() - time(); totDegTime := time() - time(); uniGcdTime := time() - time(); CMBBSHLTime := time() - time(); buildTime := time() - time();
checkTime := time() - time(); gcdPTime := time() - time(); sqfrTime := time() - time(); chremTime := time() - time(); expandTime := time() - time();

end proc;


local figureOut := time() - time();
local specificEvals;
local specificTime;

MHLBBGCDp := proc(BA,BB,X,da,db,dg,p, pponly::integer)

   local n,x,y,a1,b1,g,g1,g2,a,b,ra,alpha,i,j,k,alpha2,multip;
   local a2,b2,beta,galpha,abaralpha,bbaralpha,abar,bbar;
   local BC,BD,cg,q,ra2,gamma,aq,bq,ge,beta1,beta2,cc,dd;
   local factors,LAbar,LG,Lshared,r,r1,r2,r3,counter,Sdf1,LC1,ABAR,delta,da2,db2,dg2,ca,st,st3,st4;
   local A1,XXX,A2,A3,GEvalPartial,IND,Vars,indexes,ss;


   n := nops(X);
   x := X[1];
   ra := rand(1..p-1);

   st4 := time();

   #print(dg);

   if dg[1] = 0 then



   #print("in here");

      k := 1; while dg[k] = 0 and k < n do k:=k+1; od;
      A1 := [seq(ra(),i=1..k-1)];
      XXX := op(A1);

      BC := proc(Y,p2)

           return BA([XXX,seq(Y[i],i=1..n-k+1)],p2);

      end proc:

      BD := proc(Y,p2)

         return BB([XXX,seq(Y[i],i=1..n-k+1)],p2);
      end proc:

      return MHLBBGCDp(BC,BD,X[k..n],da[k..n],db[k..n],dg[k..n],tdega,tdegb,p);
    fi;

   alpha := Array(1..n);

   for i from 1 to n do alpha[i] := ra(); od; #print(alpha);

  #print("here is the problem"); print(da); print(n); print(alpha); print(X);

   st := time();

   #print("about to enter interp 1");
   a1 := Interp1var( BA, X, alpha, da, n, 1, p);
   #print(a1);
   b1 := Interp1var( BB, X, alpha, db, n, 1, p);
   #print(b1);
   g1 := Gcd(a1,b1,'abar','bbar') mod p;

   #print(g1);

   sqfrTime += time() - st;

   #Base Case recursion
   if n = 1 then return [g1]; fi;

   y := X[2];

   #print("before getFactors");

   #print(abar);
   #print(g1);

  #factors,LAbar,LG,Lshared := getFactors(abar,g1,x,p);

  st := time();
  r,factors,multip := UNPACK(Sqrfree(g1) mod p);
  sqfrTime += time() - st;


   Sdf1 := add(degree(factors[i]),i=1..r); #print(Sdf1);

     #printf("about to enter CMBBSHL\n");

     st3 := time();
       factors := CMBBSHL( BA,BB, factors, n, X, alpha, da,db, dg, Sdf1, n, p, 0,0,0,[1,1,1,1], 0 );
    CMBBSHLTime += time() - st3;

    #printf("n=%d\n",n);
    #print(evalCounter - specificEvals); specificEvals := evalCounter;
    #print(bivEvalInterpTime - specificTime); specificTime := bivEvalInterpTime;


    #print(time()-st3);

    #print(factors);

    # printf("finished CMBBSHL\n");

     r := nops(factors);

     st := time();
     #make factors monic;
     for i from 1 to r do
        LC1 := lcoeff(factors[i],X);
        factors[i] := factors[i]/LC1 mod p;
     od;


  g := mul(factors[i]^multip[i],i=1..r);
  buildTime += time() - st;

  #print(g); #print(ABAR);
  #print(g);  print(mods(-83*g,p)); print(p);

  if pponly = 1 then return [g]; fi;

  k := 2; while dg[k]-degree(g,X[k])=0 and k <n do k++; od;

  da2 := [seq(da[i]-degree(g,X[i]),i=k..n)];
  db2 := [seq(db[i]-degree(g,X[i]),i=k..n)];
  dg2 := [seq(dg[i]-degree(g,X[i]),i=k..n)];

  #print(Factor(g) mod p);
  #print(g); print(p);
  #print(nops(dg2));

  #print(k);

  #print(da2,db2,dg2);

  if add(dg2)=0 then return [g]; fi;

     A1 := [seq(ra(),i=1..k-1)];
     A2 := [seq(X[i]=A1[i],i=1..k-1)]; #print(A2);
     A3 := op(A1);

     GEvalPartial := Eval(g,A2) mod p;

       IND := indets(GEvalPartial);
       Vars := [seq(0,1..nops(IND))];
       indexes := [seq(0,1..nops(IND))];

 j := 1; ss := 1;
  for i from k to n do
     if member(X[i],IND) then
        Vars[j] := X[i]; indexes[j] := i;
        j := j+1;
     fi;
  od;

    #print(Factor(g) mod p); print(g); print(p);

    #print(GEvalPartial);

     #print(k); print(IND); print(Vars); print(indexes);

    #
    #print(GEvalPartial);
    BC := proc(Y,q)

        local eg;

        #print(Y);

        #print([op(point),seq(Y[indexes[i]-k+1],i=ss..nops(Vars))]);

        #print(GEvalPartial); print(Vars); print([seq(Y[indexes[i]-k+1],i=1..nops(Vars))]);

        eg := Eval64(GEvalPartial,Vars,[seq(Y[indexes[i]-k+1],i=1..nops(Vars))],q);

       #print(eg);

        if eg = 0 then return FAIL; fi;

        return BA([A3,seq(Y[i],i=1..n-k+1)],q)/eg mod q;

     end proc:

   BD := proc(Y,q)

     local eg;

      #eg := Eval(GEvalPartial,[seq(X[i]=Y[i-k+1],i=k..n)]) mod q;
     eg := Eval64(GEvalPartial,Vars,[seq(Y[indexes[i]-k+1],i=1..nops(Vars))],q);

      if eg = 0 then return FAIL; fi;

      #for i from k to n do A1[i] := Y[i-k+1]; od;

      return BB([A3,seq(Y[i],i=1..n-k+1)],q)/eg mod q;

   end proc:


    #printf("recursive call down\n");
    #Recursive call down



    #printf("test degs here\n"); print(dg); print(seq(degree(g,X[i]),i=1..n));

    #print(X[2..n]);
    #print(da2); print(db2); print(dg2);
    #print(tdega-degree(g)-degree(ABAR)); print(tdegb-degree(g));

    figureOut += time() - st4;

    #recursive call down
    #if add(dg2) > 0 then
       cg:= MHLBBGCDp(BC,BD,X[k..n],da2,db2,dg2,p,pponly);
        #print(cg);
        st := time();
        #g := Expand(cg*g) mod p;
        #g := cg*g;
        expandTime += time() - st;
        #ABAR := Expand(ca*ABAR);
    #fi;

    #printf("out of recursive call\n");

   #print(cg); #return -1;

       return [g,op(cg)];

end proc:


BBMGCD := proc(BA,BB,X, pponly::integer)

   local i,j,k,gp,g,G,Gprime,gprime,p,q,M,n,da,db,a,b,ra,x1,alpha;
   local compare,LM,ra2,gamma,ge,aq,bq,T,x,dg,tdega,tdegb,abar,st,parts,m;


   M,G := 0,0;
   n := nops(X); p := 2^61;

   if n<1 then
      return "You need at least 1 variable to use this algorithm";
   fi; #printf("got in");

   x := X[1];

   da := [seq(0,i=1..n)]; db := [seq(0,i=1..n)]; dg := [seq(0,i=1..n)]; q := nextprime(2^60); ra := rand(1..q-1);
   alpha := [seq(0,i=1..n)]; for i from 1 to n do alpha[i] := ra(); od;

   #print("before total");

   st := time();

      #print("before unis");

   #print(Factor(BA(X,q)) mod q);
   #print(Factor(BB(X,q)) mod q);

   for i from 1 to n do
       a := UnivInterpNew(BA,X,alpha,i,q); #print(Factor(Expand(a) mod q) mod q);
       b := UnivInterpNew(BB,X,alpha,i,q); #print(Factor(Expand(b) mod q) mod q);
       da[i] := degree(a);
       db[i] := degree(b);
       #print(Gcd(a,b) mod q);
       dg[i] := degree( Gcd(a,b) mod q);
   od;
   uniDegTime := uniDegTime + time() - st;

   if pponly=1 and dg[1] = 0 then return 1; fi;

   #print("got here");
   #print(da); print(db);
   #print(dg);
   #print(tdega); print(tdegb);

   st := time();
   #tdega := TotalDegNew(BA,X,q);
   #tdegb := TotalDegNew(BB,X,q);
   totDegTime := totDegTime + time() - st;

   LM := 0;
   T := 0;

   specificEvals := 0;
   specificTime := time() - time();

   while true do

      p := nextprime(p); #if p > 1030 then return -1 fi;

      st := time();
      parts := MHLBBGCDp(BA,BB,X,da,db,dg,p,pponly);
      gp := parts[1];
      m := nops(parts);
      #print(gp); print(parts);
      gcdPTime := gcdPTime + time() - st;

      if gp <> FAIL then

      st := time();
      #gp := Expand(gp) mod p;
      expandTime += time() - st;

         compare := 0;
         if LM <> 0 then
            lcoeff(gp,X,'LM2'); #print(LM2);
            if LM <> LM2 then
               #for i from 1 to n do
                for i from n to 1 by -1 do
                  k := degree(LM2,X[i]) - degree(LM,X[i]);
                  #if degree(LM2,X[i])>0 and degree(LM,X[i])=0 or degree(LM2,X[i]) < degree(LM,X[i]) then
                   if k < 0 then
                     compare := 1; break;
                  #elif degree(LM2,X[i]) > degree(LM,X[i]) then
                   elif k > 0 then
                     compare := 2; break;
                  fi;
               od;
            fi;
         fi;

         #First prime
         if G=0 then
            G := [seq(parts[i],i=1..m)];
            Gprime := [seq(iratrecon(parts[i],p),i=1..m)];
            for i from 1 to m do
               if Gprime[i] <> FAIL then Gprime[i] := denom(Gprime[i])*Gprime[i]; fi;
            od;
             lcoeff(Gprime[1],X,'LM');
             M := p;
            #gprime := iratrecon(gp,p);
            #if gprime[1] <> FAIL then gprime := denom(gprime)*gprime; fi;
            #M,G,Gprime := p,gp,gprime;

         elif compare = 1 then
            G := [seq(parts[i],i=1..m)];
            Gprime := [seq(iratrecon(parts[i],p),i=1..m)];
            for i from 1 to m do
               if Gprime[i] <> FAIL then Gprime[i] := denom(Gprime[i])*Gprime[i]; fi;
            od;
             lcoeff(Gprime[1],X,'LM');
             M := p;
         elif compare = 0 then

            #print("normal case");
           # print(gp); print(G);

          g := [seq(0,i=1..m)];
          gprime := [seq(0,i=1..m)];

           st := time();
           for i from 1 to m do
              g[i] := chrem([G[i],parts[i]],[M,p]);
              gprime[i] :=iratrecon(g[i],M*p);
              if gprime[i] <> FAIL then gprime[i] := denom(gprime[i])*gprime[i]; fi;
           od;
            #g := chrem([G,gp],[M,p]); #print(g);
            M := M*p;
           #gprime := iratrecon(g,M);

            #if gprime <> FAIL then gprime := denom(gprime)*gprime; fi;
            chremTime := chremTime + time()-st;

            if gprime[1] <> FAIL and gprime[1] = Gprime[1] then

            #print(gprime); print(m);

                  q := nextprime(12345678900); ra2 := rand(1..q-1);
                  do
                      gamma := [seq(ra2(),i=2..n)]; ge := 1;
                      for j from 1 to m do
                         ge := ge*(Eval(gprime[j],{seq(X[i]=gamma[i-1],i=2..n)}) mod q);
                      od;
                      #print(ge); print(add(degree(gprime[k],X[1]),k=1..m));
                 until degree(ge,X[1]) = add(degree(gprime[k],X[1]),k=1..m);
                 st := time();
                 aq := ProbeInterpUni2(BA,X,[ra2(),op(gamma)],X[1],1,q);
                 bq := ProbeInterpUni2(BB,X,[ra2(),op(gamma)],X[1],1,q);
                  if Divide(aq,ge) mod q and Divide(bq,ge) mod q then
                     checkTime := checkTime + time() - st;
                    return mul(gprime);
                 fi;
            fi;

            G,Gprime := g,gprime;

         elif compare = 2 then
            #print("ignore");
            #unlucky prime. do nothing

         fi;


      fi;

   od;


end proc:

printTimings := proc();

print("modular timings");
printf("Univ Interp + Gcd time=%8.2fs\n",(uniGcdTime));
printf("square-free time=%8.2fs\n",(sqfrTime));
printf("CMBBSHL time=%8.2fs\n",(CMBBSHLTime));
printf("expand gp time=%8.2fs\n",(expandTime));
printf("Build time=%8.2fs\n",(buildTime));

print("main timings");
printf("uni deg time=%8.2fs\n",(uniDegTime));
printf("tot deg time=%8.2fs\n",(totDegTime));
printf("gcd mod p time=%8.2fs\n",(gcdPTime));
printf("chrem + iratrecon time=%8.2fs\n",(chremTime));
printf("check time=%8.2fs\n",(checkTime));


end proc;
